# Copyright 2018-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.repository.storage.interface import (
    RepoStorageException,
    RepoStorageInterface,
)


class InplaceRepoStorage(RepoStorageInterface):
    """
    Legacy repo storage behavior, where updates are applied in-place.
    This module is not recommended, since the repository is left in an
    unspecified (possibly malicious) state if the update fails.
    """

    def __init__(self, repo, spawn_kwargs):
        self._user_location = repo.location
        self._update_location = None

    async def init_update(self):
        self._update_location = self._user_location
        return self._update_location

    @property
    def current_update(self):
        if self._update_location is None:
            raise RepoStorageException("current update does not exist")
        return self._update_location

    async def commit_update(self):
        self.current_update
        self._update_location = None

    async def abort_update(self):
        self._update_location = None

    async def garbage_collection(self):
        pass



# Copyright 2022-2024 Gentoo Authors
# Portage Unit Testing Functionality

import io
import tarfile
import tempfile
from functools import partial
from os import urandom
from concurrent.futures import Future

from portage.gpkg import gpkg
from portage import os
from portage import shutil
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.exception import InvalidSignature
from portage.gpg import GPG


class test_gpkg_metadata_url_case(TestCase):
    def httpd(self, directory, httpd_future):
        try:
            import http.server
            import socketserver
        except ImportError:
            self.skipTest("http server not exits")

        Handler = partial(http.server.SimpleHTTPRequestHandler, directory=directory)

        with socketserver.TCPServer(("127.0.0.1", 0), Handler) as httpd:
            httpd_future.set_result(httpd)
            httpd.serve_forever()

    def start_http_server(self, directory):
        try:
            import threading
        except ImportError:
            self.skipTest("threading module not exists")

        httpd_future = Future()
        server = threading.Thread(
            target=self.httpd, args=(directory, httpd_future), daemon=True
        )
        server.start()
        return httpd_future.result()

    def test_gpkg_get_metadata_url(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'BINPKG_COMPRESS="gzip"',
                    'FEATURES="${FEATURES} -binpkg-signing '
                    '-binpkg-request-signature"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()
        server = None
        try:
            settings = playground.settings
            server = self.start_http_server(tmpdir)

            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            with open(os.path.join(orig_full_path, "test"), "wb") as test_file:
                test_file.write(urandom(1048576))

            gpkg_file_loc = os.path.join(tmpdir, "test.gpkg.tar")
            test_gpkg = gpkg(settings, "test", gpkg_file_loc)

            meta = {
                "test1": b"{abcdefghijklmnopqrstuvwxyz, 1234567890}",
                "test2": urandom(102400),
            }

            test_gpkg.compress(os.path.join(tmpdir, "orig"), meta)

            meta_from_url = test_gpkg.get_metadata_url(
                "http://{}:{}/test.gpkg.tar".format(*server.server_address)
            )

            self.assertEqual(meta, meta_from_url)
        finally:
            if server is not None:
                server.shutdown()
            shutil.rmtree(tmpdir)
            playground.cleanup()

    def test_gpkg_get_metadata_url_unknown_signature(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'BINPKG_COMPRESS="gzip"',
                    'FEATURES="${FEATURES} binpkg-signing ' 'binpkg-request-signature"',
                ),
            }
        )
        tmpdir = tempfile.mkdtemp()
        gpg = None
        server = None
        try:
            settings = playground.settings
            gpg = GPG(settings)
            gpg.unlock()

            server = self.start_http_server(tmpdir)

            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)

            with open(os.path.join(orig_full_path, "test"), "wb") as test_file:
                test_file.write(urandom(1048576))

            gpkg_file_loc = os.path.join(tmpdir, "test-1.gpkg.tar")
            test_gpkg = gpkg(settings, "test", gpkg_file_loc)

            meta = {
                "test1": b"{abcdefghijklmnopqrstuvwxyz, 1234567890}",
                "test2": urandom(102400),
            }

            test_gpkg.compress(os.path.join(tmpdir, "orig"), meta)

            with tarfile.open(os.path.join(tmpdir, "test-1.gpkg.tar"), "r") as tar_1:
                with tarfile.open(
                    os.path.join(tmpdir, "test-2.gpkg.tar"), "w"
                ) as tar_2:
                    for f in tar_1.getmembers():
                        if f.name == "test/metadata.tar.gz":
                            sig = b"""
-----BEGIN PGP SIGNATURE-----

iHUEABYIAB0WIQRVhCbPGi/rhGTq4nV+k2dcK9uyIgUCXw4ehAAKCRB+k2dcK9uy
IkCfAP49AOYjzuQPP0n5P0SGCINnAVEXN7QLQ4PurY/lt7cT2gEAq01stXjFhrz5
87Koh+ND2r5XfQsz3XeBqbb/BpmbEgo=
=sc5K
-----END PGP SIGNATURE-----
"""
                            data = io.BytesIO(sig)
                            f.size = len(sig)
                            tar_2.addfile(f, data)
                            data.close()
                        else:
                            tar_2.addfile(f, tar_1.extractfile(f))

            test_gpkg = gpkg(settings, "test")
            self.assertRaises(
                InvalidSignature,
                test_gpkg.get_metadata_url,
                "http://{}:{}/test-2.gpkg.tar".format(*server.server_address),
            )
        finally:
            if gpg is not None:
                gpg.stop()
            if server is not None:
                server.shutdown()
            shutil.rmtree(tmpdir)
            playground.cleanup()



# Copyright 2005-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import portage
from portage import os


class WorldHandler:
    short_desc = "Fix problems in the world file"

    @staticmethod
    def name():
        return "world"

    def __init__(self):
        self.invalid = []
        self.not_installed = []
        self.okay = []
        from portage._sets import load_default_config

        setconfig = load_default_config(
            portage.settings, portage.db[portage.settings["EROOT"]]
        )
        self._sets = setconfig.getSets()

    def _check_world(self, onProgress):
        eroot = portage.settings["EROOT"]
        self.world_file = os.path.join(eroot, portage.const.WORLD_FILE)
        self.found = os.access(self.world_file, os.R_OK)
        vardb = portage.db[eroot]["vartree"].dbapi

        from portage._sets import SETPREFIX

        sets = self._sets
        world_atoms = list(sets["selected"])
        maxval = len(world_atoms)
        if onProgress:
            onProgress(maxval, 0)
        for i, atom in enumerate(world_atoms):
            if not isinstance(atom, portage.dep.Atom):
                if atom.startswith(SETPREFIX):
                    s = atom[len(SETPREFIX) :]
                    if s in sets:
                        self.okay.append(atom)
                    else:
                        self.not_installed.append(atom)
                else:
                    self.invalid.append(atom)
                if onProgress:
                    onProgress(maxval, i + 1)
                continue
            okay = True
            if not vardb.match(atom):
                self.not_installed.append(atom)
                okay = False
            if okay:
                self.okay.append(atom)
            if onProgress:
                onProgress(maxval, i + 1)

    def check(self, **kwargs):
        onProgress = kwargs.get("onProgress", None)
        self._check_world(onProgress)
        errors = []
        if self.found:
            errors += [f"'{x}' is not a valid atom" for x in self.invalid]
            errors += [f"'{x}' is not installed" for x in self.not_installed]
        else:
            errors.append(self.world_file + " could not be opened for reading")
        if errors:
            return (False, errors)
        return (True, None)

    def fix(self, **kwargs):
        onProgress = kwargs.get("onProgress", None)
        world_set = self._sets["selected"]
        world_set.lock()
        try:
            world_set.load()  # maybe it's changed on disk
            before = set(world_set)
            self._check_world(onProgress)
            after = set(self.okay)
            errors = []
            if before != after:
                try:
                    world_set.replace(self.okay)
                except portage.exception.PortageException:
                    errors.append(f"{self.world_file} could not be opened for writing")
            if errors:
                return (False, errors)
            return (True, None)
        finally:
            world_set.unlock()



# Copyright 2007-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.exception import InvalidData
from portage.localization import _
from portage._sets.base import PackageSet
from portage._sets import get_boolean, SetConfigError
import shlex


class LibraryConsumerSet(PackageSet):
    _operations = ["merge", "unmerge"]

    def __init__(self, vardbapi, debug=False):
        super().__init__()
        self.dbapi = vardbapi
        self.debug = debug

    def mapPathsToAtoms(self, paths):
        rValue = set()
        for p in paths:
            for cpv in self.dbapi._linkmap.getOwners(p):
                try:
                    pkg = self.dbapi._pkg_str(cpv, None)
                except (KeyError, InvalidData):
                    # This is expected for preserved libraries
                    # of packages that have been uninstalled
                    # without replacement.
                    pass
                else:
                    rValue.add(f"{pkg.cp}:{pkg.slot}")
        return rValue


class LibraryFileConsumerSet(LibraryConsumerSet):
    """
    Note: This does not detect libtool archive (*.la) files that consume the
    specified files (revdep-rebuild is able to detect them).
    """

    description = (
        "Package set which contains all packages "
        + "that consume the specified library file(s)."
    )

    def __init__(self, vardbapi, files, **kargs):
        super().__init__(vardbapi, **kargs)
        self.files = files

    def load(self):
        consumers = set()
        for lib in self.files:
            consumers.update(self.dbapi._linkmap.findConsumers(lib, greedy=False))

        if not consumers:
            return
        self._setAtoms(self.mapPathsToAtoms(consumers))

    def singleBuilder(cls, options, settings, trees):
        files = tuple(shlex.split(options.get("files", "")))
        if not files:
            raise SetConfigError(_("no files given"))
        debug = get_boolean(options, "debug", False)
        return LibraryFileConsumerSet(trees["vartree"].dbapi, files, debug=debug)

    singleBuilder = classmethod(singleBuilder)


class PreservedLibraryConsumerSet(LibraryConsumerSet):
    def load(self):
        reg = self.dbapi._plib_registry
        if reg is None:
            # preserve-libs is entirely disabled
            return
        consumers = set()
        if reg:
            plib_dict = reg.getPreservedLibs()
            for libs in plib_dict.values():
                for lib in libs:
                    if self.debug:
                        print(lib)
                        for x in sorted(
                            self.dbapi._linkmap.findConsumers(lib, greedy=False)
                        ):
                            print("    ", x)
                        print("-" * 40)
                    consumers.update(
                        self.dbapi._linkmap.findConsumers(lib, greedy=False)
                    )
            # Don't rebuild packages just because they contain preserved
            # libs that happen to be consumers of other preserved libs.
            for libs in plib_dict.values():
                consumers.difference_update(libs)
        else:
            return
        if not consumers:
            return
        self._setAtoms(self.mapPathsToAtoms(consumers))

    def singleBuilder(cls, options, settings, trees):
        debug = get_boolean(options, "debug", False)
        return PreservedLibraryConsumerSet(trees["vartree"].dbapi, debug=debug)

    singleBuilder = classmethod(singleBuilder)



# Copyright 2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class TarMergeOrderTestCase(TestCase):
    def testTarMergeOrder(self):
        """
        Test for bug #922629 where binary app-arch/tar[acl] was merged
        before its dependency sys-apps/acl (with virtual/acl merged but
        unsatisfied).

        It poorly interacted with @system containing app-alternatives/tar
        as a circular dependency on app-arch/tar.

        Bisect found that commit 49e01d041c74680a81860b819daff812d83df02f
        triggered the issue.
        """

        ebuilds = {
            "app-alternatives/tar-0-1": {
                "EAPI": "8",
                "RDEPEND": """
                    !<app-arch/tar-1.34-r2
                    gnu? ( >=app-arch/tar-1.34-r2 )
                    libarchive? ( app-arch/libarchive )
                """,
                "IUSE": "+gnu libarchive",
                "REQUIRED_USE": "^^ ( gnu libarchive )",
            },
            "app-arch/libarchive-3.7.4": {"EAPI": "8"},
            "app-arch/tar-1.35": {
                "EAPI": "8",
                "RDEPEND": """
                    acl? ( virtual/acl )
                """,
                "DEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "BDEPEND": """
                    nls? ( sys-devel/gettext )
                """,
                "IUSE": "acl nls xattr",
            },
            "virtual/acl-0-r2": {
                "EAPI": "8",
                "RDEPEND": ">=sys-apps/acl-2.2.52-r1",
            },
            "sys-devel/gettext-0.22.4": {
                "EAPI": "8",
                "RDEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "DEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "IUSE": "acl nls xattr",
            },
            "sys-apps/attr-2.5.2-r1": {
                "EAPI": "8",
                "BDEPEND": "nls? ( sys-devel/gettext )",
                "IUSE": "nls",
            },
            "sys-apps/acl-2.3.2-r1": {
                "EAPI": "8",
                "DEPEND": ">=sys-apps/attr-2.4.47-r1",
                "RDEPEND": ">=sys-apps/attr-2.4.47-r1",
                "BDEPEND": "nls? ( sys-devel/gettext )",
                "IUSE": "nls",
            },
        }

        installed = {
            "app-alternatives/tar-0-1": {
                "EAPI": "8",
                "RDEPEND": """
                    !<app-arch/tar-1.34-r2
                    gnu? ( >=app-arch/tar-1.34-r2 )
                    libarchive? ( app-arch/libarchive )
                """,
                "IUSE": "+gnu libarchive",
                "USE": "gnu",
                "REQUIRED_USE": "^^ ( gnu libarchive )",
            },
            "app-arch/tar-1.35": {
                "EAPI": "8",
                "RDEPEND": """
                    acl? ( virtual/acl )
                """,
                "DEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "BDEPEND": """
                    nls? ( sys-devel/gettext )
                """,
                "IUSE": "acl nls xattr",
                "USE": "",
            },
            "sys-devel/gettext-0.22.4": {
                "EAPI": "8",
                "RDEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "DEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "IUSE": "acl nls xattr",
                "USE": "xattr",
            },
            "sys-apps/attr-2.5.2-r1": {
                "EAPI": "8",
                "BDEPEND": "nls? ( sys-devel/gettext )",
                "IUSE": "nls",
                "USE": "",
            },
        }

        binpkgs = {
            "app-alternatives/tar-0-1": {
                "EAPI": "8",
                "RDEPEND": """
                    !<app-arch/tar-1.34-r2
                    gnu? ( >=app-arch/tar-1.34-r2 )
                    libarchive? ( app-arch/libarchive )
                """,
                "IUSE": "+gnu libarchive",
                "USE": "gnu",
                "REQUIRED_USE": "^^ ( gnu libarchive )",
            },
            "app-arch/tar-1.35": {
                "EAPI": "8",
                "RDEPEND": """
                    acl? ( virtual/acl )
                """,
                "DEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "BDEPEND": """
                    nls? ( sys-devel/gettext )
                """,
                "IUSE": "acl nls xattr",
                "USE": "acl nls xattr",
            },
            "virtual/acl-0-r2": {
                "EAPI": "8",
                "RDEPEND": ">=sys-apps/acl-2.2.52-r1",
            },
            "sys-devel/gettext-0.22.4": {
                "EAPI": "8",
                "RDEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "DEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "IUSE": "acl nls xattr",
                "USE": "acl nls xattr",
            },
            "sys-apps/attr-2.5.2-r1": {
                "EAPI": "8",
                "BDEPEND": "nls? ( sys-devel/gettext )",
                "IUSE": "nls",
                "USE": "nls",
            },
        }

        world = []

        user_config = {
            "package.use": (
                "app-arch/tar acl nls xattr",
                "sys-apps/acl nls",
                "sys-apps/attr nls",
                "sys-devel/gettext acl nls xattr",
            ),
        }

        playground = ResolverPlayground(
            ebuilds=ebuilds,
            installed=installed,
            binpkgs=binpkgs,
            world=world,
            user_config=user_config,
        )
        settings = playground.settings
        profile_path = settings.profile_path

        # Add app-alternatives/tar to @system too
        with open(os.path.join(profile_path, "packages"), "w") as f:
            f.writelines(["*app-alternatives/tar\n", "*app-arch/tar\n"])
        test_cases = (
            # Check without binpkgs first
            ResolverPlaygroundTestCase(
                ["@world"],
                success=True,
                options={"--emptytree": True},
                mergelist=[
                    "sys-apps/acl-2.3.2-r1",
                    "virtual/acl-0-r2",
                    "sys-apps/attr-2.5.2-r1",
                    "sys-devel/gettext-0.22.4",
                    "app-arch/tar-1.35",
                    "app-alternatives/tar-0",
                ],
            ),
            # In the bug, only --emptytree was broken, so check
            # some cases without it.
            ResolverPlaygroundTestCase(
                ["@world"],
                success=True,
                options={
                    "--usepkg": True,
                },
                mergelist=[
                    "sys-apps/acl-2.3.2-r1",
                    "[binary]virtual/acl-0-r2",
                    "[binary]app-arch/tar-1.35",
                    "[binary]app-alternatives/tar-0",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["app-arch/tar"],
                success=True,
                options={
                    "--oneshot": True,
                    "--usepkg": True,
                },
                mergelist=[
                    "sys-apps/acl-2.3.2-r1",
                    "[binary]virtual/acl-0-r2",
                    "[binary]app-arch/tar-1.35",
                ],
            ),
            # binpkg --emptytree case which broke
            ResolverPlaygroundTestCase(
                ["@world"],
                success=True,
                options={
                    "--emptytree": True,
                    "--usepkg": True,
                },
                mergelist=[
                    "[binary]sys-apps/attr-2.5.2-r1",
                    "[binary]virtual/acl-0-r2",
                    "[binary]sys-devel/gettext-0.22.4",
                    "sys-apps/acl-2.3.2-r1",
                    "[binary]app-arch/tar-1.35",
                    "[binary]app-alternatives/tar-0",
                ],
            ),
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testTarMergeOrderWithoutAlternatives(self):
        """
        Variant of test for bug #922629 where binary app-arch/tar[acl] was merged
        before its dependency sys-apps/acl (with virtual/acl merged but
        unsatisfied).

        This variant lacks the problematic app-alternatives/tar to check we handle
        the simpler case correctly.
        """

        ebuilds = {
            "app-arch/tar-1.35": {
                "EAPI": "8",
                "RDEPEND": """
                    acl? ( virtual/acl )
                """,
                "DEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "BDEPEND": """
                    nls? ( sys-devel/gettext )
                """,
                "IUSE": "acl nls xattr",
            },
            "virtual/acl-0-r2": {
                "EAPI": "8",
                "RDEPEND": ">=sys-apps/acl-2.2.52-r1",
            },
            "sys-devel/gettext-0.22.4": {
                "EAPI": "8",
                "RDEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "DEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "IUSE": "acl nls xattr",
            },
            "sys-apps/attr-2.5.2-r1": {
                "EAPI": "8",
                "BDEPEND": "nls? ( sys-devel/gettext )",
                "IUSE": "nls",
            },
            "sys-apps/acl-2.3.2-r1": {
                "EAPI": "8",
                "DEPEND": ">=sys-apps/attr-2.4.47-r1",
                "RDEPEND": ">=sys-apps/attr-2.4.47-r1",
                "BDEPEND": "nls? ( sys-devel/gettext )",
                "IUSE": "nls",
            },
        }

        installed = {
            "app-arch/tar-1.35": {
                "EAPI": "8",
                "RDEPEND": """
                    acl? ( virtual/acl )
                """,
                "DEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "BDEPEND": """
                    nls? ( sys-devel/gettext )
                """,
                "IUSE": "acl nls xattr",
                "USE": "",
            },
            "sys-devel/gettext-0.22.4": {
                "EAPI": "8",
                "RDEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "DEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "IUSE": "acl nls xattr",
                "USE": "xattr",
            },
            "sys-apps/attr-2.5.2-r1": {
                "EAPI": "8",
                "BDEPEND": "nls? ( sys-devel/gettext )",
                "IUSE": "nls",
                "USE": "",
            },
        }

        binpkgs = {
            "app-arch/tar-1.35": {
                "EAPI": "8",
                "RDEPEND": """
                    acl? ( virtual/acl )
                """,
                "DEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "BDEPEND": """
                    nls? ( sys-devel/gettext )
                """,
                "IUSE": "acl nls xattr",
                "USE": "acl nls xattr",
            },
            "virtual/acl-0-r2": {
                "EAPI": "8",
                "RDEPEND": ">=sys-apps/acl-2.2.52-r1",
            },
            "sys-devel/gettext-0.22.4": {
                "EAPI": "8",
                "RDEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "DEPEND": """
                    acl? ( virtual/acl )
                    xattr? ( sys-apps/attr )
                """,
                "IUSE": "acl nls xattr",
                "USE": "acl nls xattr",
            },
            "sys-apps/attr-2.5.2-r1": {
                "EAPI": "8",
                "BDEPEND": "nls? ( sys-devel/gettext )",
                "IUSE": "nls",
                "USE": "nls",
            },
        }

        world = []

        user_config = {
            "package.use": (
                "app-arch/tar acl nls xattr",
                "sys-apps/acl nls",
                "sys-apps/attr nls",
                "sys-devel/gettext acl nls xattr",
            ),
        }

        playground = ResolverPlayground(
            ebuilds=ebuilds,
            installed=installed,
            binpkgs=binpkgs,
            world=world,
            user_config=user_config,
        )
        settings = playground.settings
        profile_path = settings.profile_path

        with open(os.path.join(profile_path, "packages"), "w") as f:
            f.writelines(["*app-arch/tar\n"])
        test_cases = (
            # Check without binpkgs first
            ResolverPlaygroundTestCase(
                ["@world"],
                success=True,
                options={"--emptytree": True, "--verbose": True},
                mergelist=[
                    "sys-apps/acl-2.3.2-r1",
                    "virtual/acl-0-r2",
                    "sys-apps/attr-2.5.2-r1",
                    "sys-devel/gettext-0.22.4",
                    "app-arch/tar-1.35",
                ],
            ),
            # In the bug, only --emptytree was broken, so check
            # some cases without it.
            ResolverPlaygroundTestCase(
                ["@world"],
                success=True,
                options={
                    "--usepkg": True,
                },
                mergelist=[
                    "sys-apps/acl-2.3.2-r1",
                    "[binary]virtual/acl-0-r2",
                    "[binary]app-arch/tar-1.35",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["app-arch/tar"],
                success=True,
                options={
                    "--oneshot": True,
                    "--usepkg": True,
                },
                mergelist=[
                    "sys-apps/acl-2.3.2-r1",
                    "[binary]virtual/acl-0-r2",
                    "[binary]app-arch/tar-1.35",
                ],
            ),
            # binpkg --emptytree case which broke
            ResolverPlaygroundTestCase(
                ["@world"],
                success=True,
                options={
                    "--emptytree": True,
                    "--usepkg": True,
                },
                mergelist=[
                    "[binary]sys-apps/attr-2.5.2-r1",
                    "[binary]virtual/acl-0-r2",
                    "[binary]sys-devel/gettext-0.22.4",
                    "sys-apps/acl-2.3.2-r1",
                    "[binary]app-arch/tar-1.35",
                ],
            ),
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# test_best_match_to_list.py -- Portage Unit Testing Functionality
# Copyright 2010-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from itertools import permutations

from portage.tests import TestCase
from portage.dep import Atom, best_match_to_list


class Test_best_match_to_list(TestCase):
    def best_match_to_list_wrapper(self, mypkg, mylist):
        """
        This function uses best_match_to_list to create sorted
        list of matching atoms.
        """
        ret = []
        mylist = list(mylist)
        while mylist:
            m = best_match_to_list(mypkg, mylist)
            if m is not None:
                ret.append(m)
                mylist.remove(m)
            else:
                break

        return ret

    def testBest_match_to_list(self):
        tests = [
            (
                "dev-libs/A-4",
                [Atom(">=dev-libs/A-3"), Atom(">=dev-libs/A-2")],
                [Atom(">=dev-libs/A-3"), Atom(">=dev-libs/A-2")],
                True,
            ),
            (
                "dev-libs/A-4",
                [Atom("<=dev-libs/A-5"), Atom("<=dev-libs/A-6")],
                [Atom("<=dev-libs/A-5"), Atom("<=dev-libs/A-6")],
                True,
            ),
            (
                "dev-libs/A-1",
                [Atom("dev-libs/A"), Atom("=dev-libs/A-1")],
                [Atom("=dev-libs/A-1"), Atom("dev-libs/A")],
                True,
            ),
            (
                "dev-libs/A-1",
                [Atom("dev-libs/B"), Atom("=dev-libs/A-1:0")],
                [Atom("=dev-libs/A-1:0")],
                True,
            ),
            (
                "dev-libs/A-1",
                [Atom("dev-libs/*", allow_wildcard=True), Atom("=dev-libs/A-1:0")],
                [Atom("=dev-libs/A-1:0"), Atom("dev-libs/*", allow_wildcard=True)],
                True,
            ),
            (
                "dev-libs/A-4.9999-r1",
                [
                    Atom("dev-libs/*", allow_wildcard=True),
                    Atom("=*/*-*9999*", allow_wildcard=True),
                ],
                [
                    Atom("=*/*-*9999*", allow_wildcard=True),
                    Atom("dev-libs/*", allow_wildcard=True),
                ],
                True,
            ),
            (
                "dev-libs/A-4_beta-r1",
                [
                    Atom("dev-libs/*", allow_wildcard=True),
                    Atom("=*/*-*_beta*", allow_wildcard=True),
                ],
                [
                    Atom("=*/*-*_beta*", allow_wildcard=True),
                    Atom("dev-libs/*", allow_wildcard=True),
                ],
                True,
            ),
            (
                "dev-libs/A-4_beta1-r1",
                [
                    Atom("dev-libs/*", allow_wildcard=True),
                    Atom("=*/*-*_beta*", allow_wildcard=True),
                ],
                [
                    Atom("=*/*-*_beta*", allow_wildcard=True),
                    Atom("dev-libs/*", allow_wildcard=True),
                ],
                True,
            ),
            (
                "dev-libs/A-1:0",
                [
                    Atom("dev-*/*", allow_wildcard=True),
                    Atom("dev-*/*:0", allow_wildcard=True),
                    Atom("dev-libs/A"),
                    Atom("<=dev-libs/A-2"),
                    Atom("dev-libs/A:0"),
                    Atom("=dev-libs/A-1*"),
                    Atom("~dev-libs/A-1"),
                    Atom("=dev-libs/A-1"),
                ],
                [
                    Atom("=dev-libs/A-1"),
                    Atom("~dev-libs/A-1"),
                    Atom("=dev-libs/A-1*"),
                    Atom("dev-libs/A:0"),
                    Atom("<=dev-libs/A-2"),
                    Atom("dev-libs/A"),
                    Atom("dev-*/*:0", allow_wildcard=True),
                    Atom("dev-*/*", allow_wildcard=True),
                ],
                False,
            ),
        ]

        for pkg, atom_list, result, all_permutations in tests:
            if all_permutations:
                atom_lists = permutations(atom_list)
            else:
                atom_lists = [atom_list]
            for atom_list in atom_lists:
                self.assertEqual(
                    self.best_match_to_list_wrapper(pkg, atom_list), result
                )



# config.py -- Portage Config
# Copyright 2007-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import stat

from portage import os
from portage import _encodings
from portage import _unicode_decode
from portage import _unicode_encode
from portage.localization import _


class LoaderError(Exception):
    def __init__(self, resource, error_msg):
        """
        @param resource: Resource that failed to load (file/sql/etc)
        @type resource: String
        @param error_msg: Error from underlying Loader system
        @type error_msg: String
        """

        self.resource = resource
        self.error_msg = error_msg

    def __str__(self):
        return "Failed while loading resource: {}, error was: {}".format(
            self.resource,
            self.error_msg,
        )


def RecursiveFileLoader(filename):
    """
    If filename is of type file, return a generate that yields filename
    else if filename is of type directory, return a generator that fields
    files in that directory.

    Ignore files beginning with . or ending in ~.
    Prune CVS directories.

    @param filename: name of a file/directory to traverse
    @rtype: list
    @return: List of files to process
    """

    try:
        st = os.stat(filename)
    except OSError:
        return
    if stat.S_ISDIR(st.st_mode):
        for root, dirs, files in os.walk(filename):
            for d in list(dirs):
                if d[:1] == "." or d == "CVS":
                    dirs.remove(d)
            for f in files:
                try:
                    f = _unicode_decode(f, encoding=_encodings["fs"], errors="strict")
                except UnicodeDecodeError:
                    continue
                if f[:1] == "." or f[-1:] == "~":
                    continue
                yield os.path.join(root, f)
    else:
        yield filename


class DataLoader:
    def __init__(self, validator):
        f = validator
        if f is None:
            # if they pass in no validator, just make a fake one
            # that always returns true
            def validate(key):
                return True

            f = validate
        self._validate = f

    def load(self):
        """
        Function to do the actual work of a Loader
        """
        raise NotImplementedError("Please override in a subclass")


class EnvLoader(DataLoader):
    """Class to access data in the environment"""

    def __init__(self, validator):
        DataLoader.__init__(self, validator)

    def load(self):
        return os.environ


class TestTextLoader(DataLoader):
    """You give it some data, it 'loads' it for you, no filesystem access"""

    def __init__(self, validator):
        DataLoader.__init__(self, validator)
        self.data = {}
        self.errors = {}

    def setData(self, text):
        """Explicitly set the data field
        Args:
                text - a dict of data typical of Loaders
        Returns:
                None
        """
        if isinstance(text, dict):
            self.data = text
        else:
            raise ValueError("setData requires a dict argument")

    def setErrors(self, errors):
        self.errors = errors

    def load(self):
        return (self.data, self.errors)


class FileLoader(DataLoader):
    """Class to access data in files"""

    def __init__(self, filename, validator):
        """
        Args:
                filename : Name of file or directory to open
                validator : class with validate() method to validate data.
        """
        DataLoader.__init__(self, validator)
        self.fname = filename

    def load(self):
        """
        Return the {source: {key: value}} pairs from a file
        Return the {source: [list of errors] from a load

        @param recursive: If set and self.fname is a directory;
                load all files in self.fname
        @type: Boolean
        @rtype: tuple
        @return:
        Returns (data,errors), both may be empty dicts or populated.
        """
        from portage.util import writemsg

        data = {}
        errors = {}
        # I tried to save a nasty lookup on lineparser by doing the lookup
        # once, which may be expensive due to digging in child classes.
        func = self.lineParser
        for fn in RecursiveFileLoader(self.fname):
            try:
                with open(
                    _unicode_encode(fn, encoding=_encodings["fs"], errors="strict"),
                    encoding=_encodings["content"],
                    errors="replace",
                ) as f:
                    lines = f.readlines()
            except OSError as e:
                if e.errno == errno.EACCES:
                    writemsg(_("Permission denied: '%s'\n") % fn, noiselevel=-1)
                    del e
                elif e.errno in (errno.ENOENT, errno.ESTALE):
                    del e
                else:
                    raise
            else:
                for line_num, line in enumerate(lines):
                    func(line, line_num, data, errors)
        return (data, errors)

    def lineParser(self, line, line_num, data, errors):
        """This function parses 1 line at a time
        Args:
                line: a string representing 1 line of a file
                line_num: an integer representing what line we are processing
                data: a dict that contains the data we have extracted from the file
                      already
                errors: a dict representing parse errors.
        Returns:
                Nothing (None).  Writes to data and errors
        """
        raise NotImplementedError("Please over-ride this in a child class")


class ItemFileLoader(FileLoader):
    """
    Class to load data from a file full of items one per line

    >>> item1
    >>> item2
    >>> item3
    >>> item1

    becomes { 'item1':None, 'item2':None, 'item3':None }
    Note that due to the data store being a dict, duplicates
    are removed.
    """

    def __init__(self, filename, validator):
        FileLoader.__init__(self, filename, validator)

    def lineParser(self, line, line_num, data, errors):
        line = line.strip()
        if line.startswith("#"):  # Skip commented lines
            return
        if not len(line):  # skip empty lines
            return
        split = line.split()
        if not len(split):
            errors.setdefault(self.fname, []).append(
                _("Malformed data at line: %s, data: %s") % (line_num + 1, line)
            )
            return
        key = split[0]
        if not self._validate(key):
            errors.setdefault(self.fname, []).append(
                _("Validation failed at line: %s, data %s") % (line_num + 1, key)
            )
            return
        data[key] = None


class KeyListFileLoader(FileLoader):
    """
    Class to load data from a file full of key [list] tuples

    >>>>key foo1 foo2 foo3
    becomes
    {'key':['foo1','foo2','foo3']}
    """

    def __init__(self, filename, validator=None, valuevalidator=None):
        FileLoader.__init__(self, filename, validator)

        f = valuevalidator
        if f is None:
            # if they pass in no validator, just make a fake one
            # that always returns true
            def validate(key):
                return True

            f = validate
        self._valueValidate = f

    def lineParser(self, line, line_num, data, errors):
        line = line.strip()
        if line.startswith("#"):  # Skip commented lines
            return
        if not len(line):  # skip empty lines
            return
        split = line.split()
        if len(split) < 1:
            errors.setdefault(self.fname, []).append(
                _("Malformed data at line: %s, data: %s") % (line_num + 1, line)
            )
            return
        key = split[0]
        value = split[1:]
        if not self._validate(key):
            errors.setdefault(self.fname, []).append(
                _("Key validation failed at line: %s, data %s") % (line_num + 1, key)
            )
            return
        if not self._valueValidate(value):
            errors.setdefault(self.fname, []).append(
                _("Value validation failed at line: %s, data %s")
                % (line_num + 1, value)
            )
            return
        if key in data:
            data[key].append(value)
        else:
            data[key] = value


class KeyValuePairFileLoader(FileLoader):
    """
    Class to load data from a file full of key=value pairs

    >>>>key=value
    >>>>foo=bar
    becomes:
    {'key':'value',
     'foo':'bar'}
    """

    def __init__(self, filename, validator, valuevalidator=None):
        FileLoader.__init__(self, filename, validator)

        f = valuevalidator
        if f is None:
            # if they pass in no validator, just make a fake one
            # that always returns true
            def validate(key):
                return True

            f = validate
        self._valueValidate = f

    def lineParser(self, line, line_num, data, errors):
        line = line.strip()
        if line.startswith("#"):  # skip commented lines
            return
        if not len(line):  # skip empty lines
            return
        split = line.split("=", 1)
        if len(split) < 2:
            errors.setdefault(self.fname, []).append(
                _("Malformed data at line: %s, data %s") % (line_num + 1, line)
            )
            return
        key = split[0].strip()
        value = split[1].strip()
        if not key:
            errors.setdefault(self.fname, []).append(
                _("Malformed key at line: %s, key %s") % (line_num + 1, key)
            )
            return
        if not self._validate(key):
            errors.setdefault(self.fname, []).append(
                _("Key validation failed at line: %s, data %s") % (line_num + 1, key)
            )
            return
        if not self._valueValidate(value):
            errors.setdefault(self.fname, []).append(
                _("Value validation failed at line: %s, data %s")
                % (line_num + 1, value)
            )
            return
        data[key] = value



# Copyright 2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import portage
from portage import os
from portage.const import CACHE_PATH, PROFILE_PATH


def _get_legacy_global(name):
    constructed = portage._legacy_globals_constructed
    if name in constructed:
        return getattr(portage, name)

    if name == "portdb":
        portage.portdb = portage.db[portage.root]["porttree"].dbapi
        constructed.add(name)
        return getattr(portage, name)

    if name in ("mtimedb", "mtimedbfile"):
        portage.mtimedbfile = os.path.join(
            portage.settings["EROOT"], CACHE_PATH, "mtimedb"
        )
        constructed.add("mtimedbfile")
        portage.mtimedb = portage.MtimeDB(portage.mtimedbfile)
        constructed.add("mtimedb")
        return getattr(portage, name)

    # Portage needs to ensure a sane umask for the files it creates.
    os.umask(0o22)

    kwargs = {}
    for k, envvar in (
        ("config_root", "PORTAGE_CONFIGROOT"),
        ("target_root", "ROOT"),
        ("sysroot", "SYSROOT"),
        ("eprefix", "EPREFIX"),
    ):
        kwargs[k] = os.environ.get(envvar)

    portage._initializing_globals = True
    portage.db = portage.create_trees(**kwargs)
    constructed.add("db")
    del portage._initializing_globals

    settings = portage.db[portage.db._target_eroot]["vartree"].settings

    portage.settings = settings
    constructed.add("settings")

    # Since portage.db now uses EROOT for keys instead of ROOT, we make
    # portage.root refer to EROOT such that it continues to work as a key.
    portage.root = portage.db._target_eroot
    constructed.add("root")

    # COMPATIBILITY
    # These attributes should not be used within
    # Portage under any circumstances.

    portage.archlist = settings.archlist()
    constructed.add("archlist")

    portage.features = settings.features
    constructed.add("features")

    portage.groups = settings.get("ACCEPT_KEYWORDS", "").split()
    constructed.add("groups")

    portage.pkglines = settings.packages
    constructed.add("pkglines")

    portage.selinux_enabled = settings.selinux_enabled()
    constructed.add("selinux_enabled")

    portage.thirdpartymirrors = settings.thirdpartymirrors()
    constructed.add("thirdpartymirrors")

    profiledir = os.path.join(settings["PORTAGE_CONFIGROOT"], PROFILE_PATH)
    if not os.path.isdir(profiledir):
        profiledir = None
    portage.profiledir = profiledir
    constructed.add("profiledir")

    return getattr(portage, name)



# config.py -- Portage Config
# Copyright 2007-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = [
    "ConfigLoaderKlass",
    "GenericFile",
    "PackageKeywordsFile",
    "PackageUseFile",
    "PackageMaskFile",
    "PortageModulesFile",
]

from portage.cache.mappings import UserDict
from portage.env.loaders import (
    KeyListFileLoader,
    KeyValuePairFileLoader,
    ItemFileLoader,
)


class ConfigLoaderKlass(UserDict):
    """
    A base class stub for things to inherit from.
    Users may want a non-file backend.
    """

    def __init__(self, loader):
        """
        @param loader: A class that has a load() that returns two dicts
                the first being a data dict, the second being a dict of errors.
        """
        UserDict.__init__(self)
        self._loader = loader

    def load(self):
        """
        Load the data from the loader.

        @throws LoaderError:
        """

        self.data, self.errors = self._loader.load()


class GenericFile(UserDict):
    """
    Inherits from ConfigLoaderKlass, attempts to use all known loaders
    until it gets <something> in data.  This is probably really slow but is
    helpful when you really have no idea what you are loading (hint hint the file
    should perhaps declare  what type it is? ;)
    """

    loaders = [KeyListFileLoader, KeyValuePairFileLoader, ItemFileLoader]

    def __init__(self, filename):
        UserDict.__init__(self)
        self.filename = filename

    def load(self):
        for loader in self.loaders:
            l = loader(self.filename, None)
            data, errors = l.load()
            if len(data) and not len(errors):
                (self.data, self.errors) = (data, errors)
                return


class PackageKeywordsFile(ConfigLoaderKlass):
    """
    Inherits from ConfigLoaderKlass; implements a file-based backend.
    """

    default_loader = KeyListFileLoader

    def __init__(self, filename):
        super().__init__(self.default_loader(filename, validator=None))


class PackageUseFile(ConfigLoaderKlass):
    """
    Inherits from PackageUse; implements a file-based backend.  Doesn't handle recursion yet.
    """

    default_loader = KeyListFileLoader

    def __init__(self, filename):
        super().__init__(self.default_loader(filename, validator=None))


class PackageMaskFile(ConfigLoaderKlass):
    """
    A class that implements a file-based package.mask

    Entries in package.mask are of the form:
    atom1
    atom2
    or optionally
    -atom3
    to revert a previous mask; this only works when masking files are stacked
    """

    default_loader = ItemFileLoader

    def __init__(self, filename):
        super().__init__(self.default_loader(filename, validator=None))


class PortageModulesFile(ConfigLoaderKlass):
    """
    File Class for /etc/portage/modules
    """

    default_loader = KeyValuePairFileLoader

    def __init__(self, filename):
        super().__init__(self.default_loader(filename, validator=None))



# Copyright 1999-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from _emerge.DepPriority import DepPriority


class DepPrioritySatisfiedRange:
    """
    DepPriority                         Index      Category

    not satisfied and buildtime                    HARD
    not satisfied and runtime              7       MEDIUM
    satisfied and buildtime_slot_op        6       MEDIUM_SOFT
    satisfied and buildtime                5       MEDIUM_SOFT
    satisfied and runtime                  4       MEDIUM_SOFT
    runtime_post                           3       MEDIUM_POST
    satisfied and runtime_post             2       MEDIUM_POST
    optional                               1       SOFT
    (none of the above)                    0       NONE
    """

    MEDIUM = 7
    MEDIUM_SOFT = 6
    MEDIUM_POST = 3
    SOFT = 1
    NONE = 0

    @classmethod
    def _ignore_optional(cls, priority):
        if priority.__class__ is not DepPriority:
            return False
        return bool(priority.optional)

    @classmethod
    def _ignore_satisfied_runtime_post(cls, priority):
        if priority.__class__ is not DepPriority:
            return False
        if priority.optional:
            return True
        if not priority.satisfied:
            return False
        if priority.buildtime or priority.runtime:
            return False
        return bool(priority.runtime_post)

    @classmethod
    def _ignore_runtime_post(cls, priority):
        if priority.__class__ is not DepPriority:
            return False
        if priority.optional:
            return True
        if priority.buildtime or priority.runtime:
            return False
        return bool(priority.runtime_post)

    @classmethod
    def _ignore_satisfied_runtime(cls, priority):
        if priority.__class__ is not DepPriority:
            return False
        if priority.optional:
            return True
        if priority.buildtime:
            return False
        if not priority.runtime:
            return True
        return bool(priority.satisfied)

    @classmethod
    def _ignore_satisfied_buildtime(cls, priority):
        if priority.__class__ is not DepPriority:
            return False
        if priority.optional:
            return True
        if priority.buildtime_slot_op:
            return False
        return bool(priority.satisfied)

    @classmethod
    def _ignore_satisfied_buildtime_slot_op(cls, priority):
        if priority.__class__ is not DepPriority:
            return False
        if priority.optional:
            return True
        if priority.satisfied:
            return True
        return not priority.buildtime and not priority.runtime

    @classmethod
    def _ignore_runtime(cls, priority):
        if priority.__class__ is not DepPriority:
            return False
        # We could split this up into 2 variants (ignoring satisfied
        # runtime_slot_op, and not) if we need more granularity for ignore_priority
        # in future.
        return bool(
            (
                (not priority.runtime_slot_op)
                or (priority.satisfied and priority.runtime_slot_op)
                or priority.cross
            )
            and (priority.satisfied or priority.optional or not priority.buildtime)
        )

    ignore_medium = _ignore_runtime
    ignore_medium_soft = _ignore_satisfied_buildtime_slot_op
    ignore_medium_post = _ignore_runtime_post
    ignore_medium_post_satisifed = _ignore_satisfied_runtime_post
    ignore_soft = _ignore_optional


DepPrioritySatisfiedRange.ignore_priority = (
    None,
    DepPrioritySatisfiedRange._ignore_optional,
    DepPrioritySatisfiedRange._ignore_satisfied_runtime_post,
    DepPrioritySatisfiedRange._ignore_runtime_post,
    DepPrioritySatisfiedRange._ignore_satisfied_runtime,
    DepPrioritySatisfiedRange._ignore_satisfied_buildtime,
    DepPrioritySatisfiedRange._ignore_satisfied_buildtime_slot_op,
    DepPrioritySatisfiedRange._ignore_runtime,
)



# Copyright 2019, 2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import os
import platform

import pytest

import portage.process
from portage.const import BASH_BINARY
from portage.tests import TestCase


CLONE_NEWNET = 0x40000000
UNSHARE_NET_TEST_SCRIPT = """
ping -c 1 -W 1 127.0.0.1 || exit 1
ping -c 1 -W 1 10.0.0.1 || exit 1
[[ -n ${IPV6} ]] || exit 0
ping -c 1 -W 1 ::1 || exit 1
ping -c 1 -W 1 fd::1 || exit 1
"""


class UnshareNetTestCase(TestCase):
    def setUp(self):
        """
        Initialize ABILITY_TO_UNSHARE in setUp so that _unshare_validate
        uses the correct PORTAGE_MULTIPROCESSING_START_METHOD setup
        from super().setUp().
        """
        super().setUp()
        self.ABILITY_TO_UNSHARE = portage.process._unshare_validate(CLONE_NEWNET)

    @pytest.mark.skipif(
        portage.process.find_binary("ping") is None, reason="ping not found"
    )
    @pytest.mark.skipif(platform.system() != "Linux", reason="not Linux")
    def testUnshareNet(self):
        if self.ABILITY_TO_UNSHARE != 0:
            pytest.skip(
                f"Unable to unshare: {errno.errorcode.get(self.ABILITY_TO_UNSHARE, '?')}"
            )
        env = os.environ.copy()
        env["IPV6"] = "1" if portage.process.has_ipv6() else ""
        self.assertEqual(
            portage.process.spawn(
                [BASH_BINARY, "-c", UNSHARE_NET_TEST_SCRIPT], unshare_net=True, env=env
            ),
            0,
        )



# Copyright 2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

__all__ = ("features_set",)

import logging

from portage.const import SUPPORTED_FEATURES
from portage.localization import _
from portage.output import colorize
from portage.util import writemsg_level


class features_set:
    """
    Provides relevant set operations needed for access and modification of
    config.features. The FEATURES variable is automatically synchronized
    upon modification.

    Modifications result in a permanent override that will cause the change
    to propagate to the incremental stacking mechanism in config.regenerate().
    This eliminates the need to call config.backup_changes() when FEATURES
    is modified, since any overrides are guaranteed to persist despite calls
    to config.reset().
    """

    def __init__(self, settings):
        self._settings = settings
        self._features = set()

    def __contains__(self, k):
        return k in self._features

    def __iter__(self):
        return iter(self._features)

    def _sync_env_var(self):
        self._settings["FEATURES"] = " ".join(sorted(self._features))

    def add(self, k):
        self._settings.modifying()
        self._settings._features_overrides.append(k)
        if k not in self._features:
            self._features.add(k)
            self._sync_env_var()

    def update(self, values):
        self._settings.modifying()
        values = list(values)
        self._settings._features_overrides.extend(values)
        need_sync = False
        for k in values:
            if k in self._features:
                continue
            self._features.add(k)
            need_sync = True
        if need_sync:
            self._sync_env_var()

    def difference_update(self, values):
        self._settings.modifying()
        values = list(values)
        self._settings._features_overrides.extend("-" + k for k in values)
        remove_us = self._features.intersection(values)
        if remove_us:
            self._features.difference_update(values)
            self._sync_env_var()

    def remove(self, k):
        """
        This never raises KeyError, since it records a permanent override
        that will prevent the given flag from ever being added again by
        incremental stacking in config.regenerate().
        """
        self.discard(k)

    def discard(self, k):
        self._settings.modifying()
        self._settings._features_overrides.append("-" + k)
        if k in self._features:
            self._features.remove(k)
            self._sync_env_var()

    def _validate(self):
        """
        Implements unknown-features-warn and unknown-features-filter.
        """
        if "unknown-features-warn" in self._features:
            unknown_features = self._features.difference(SUPPORTED_FEATURES)
            if unknown_features:
                unknown_features = unknown_features.difference(
                    self._settings._unknown_features
                )
                if unknown_features:
                    self._settings._unknown_features.update(unknown_features)
                    writemsg_level(
                        colorize(
                            "BAD",
                            _("FEATURES variable contains unknown value(s): %s")
                            % ", ".join(sorted(unknown_features)),
                        )
                        + "\n",
                        level=logging.WARNING,
                        noiselevel=-1,
                    )

        if "unknown-features-filter" in self._features:
            unknown_features = self._features.difference(SUPPORTED_FEATURES)
            if unknown_features:
                self.difference_update(unknown_features)
                self._prune_overrides()

    def _prune_overrides(self):
        """
        If there are lots of invalid package.env FEATURES settings
        then unknown-features-filter can make _features_overrides
        grow larger and larger, so prune it. This performs incremental
        stacking with preservation of negative values since they need
        to persist for future config.regenerate() calls.
        """
        overrides_set = set(self._settings._features_overrides)
        positive = set()
        negative = set()
        for x in self._settings._features_overrides:
            if x[:1] == "-":
                positive.discard(x[1:])
                negative.add(x[1:])
            else:
                positive.add(x)
                negative.discard(x)
        self._settings._features_overrides[:] = list(positive) + list(
            "-" + x for x in negative
        )



# Copyright 2017-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class OnlydepsMinimalTestCase(TestCase):

    def testOnlydepsMinimal(self):
        ebuilds = {
            "dev-libs/A-1": {
                "EAPI": "8",
                "DEPEND": "dev-libs/B",
                "RDEPEND": "dev-libs/C",
                "PDEPEND": "dev-libs/D",
                "IDEPEND": "dev-libs/E",
            },
            "dev-libs/B-1": {},
            "dev-libs/C-1": {},
            "dev-libs/D-1": {},
            "dev-libs/E-1": {},
        }
        installed = {}

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                all_permutations=True,
                success=True,
                options={"--onlydeps": True, "--onlydeps-with-rdeps": "y"},
                ambiguous_merge_order=True,
                mergelist=[
                    ("dev-libs/B-1", "dev-libs/C-1", "dev-libs/D-1", "dev-libs/E-1")
                ],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                all_permutations=True,
                success=True,
                options={"--onlydeps": True, "--onlydeps-with-rdeps": "n"},
                mergelist=["dev-libs/B-1"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                all_permutations=True,
                success=True,
                options={
                    "--onlydeps": True,
                    "--onlydeps-with-rdeps": "n",
                    "--onlydeps-with-ideps": "y",
                },
                ambiguous_merge_order=True,
                mergelist=[("dev-libs/B-1", "dev-libs/E-1")],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                all_permutations=True,
                success=True,
                options={
                    "--onlydeps": True,
                    "--onlydeps-with-rdeps": "n",
                    "--onlydeps-with-ideps": "n",
                },
                mergelist=["dev-libs/B-1"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 1999-2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import logging
import shlex

import portage
from portage import os
from portage.util import writemsg_level
from portage.sync.syncbase import NewBase


class SVNSync(NewBase):
    """SVN sync module"""

    short_desc = "Perform sync operations on SVN repositories"

    @staticmethod
    def name():
        return "SVNSync"

    def __init__(self):
        NewBase.__init__(self, "svn", "dev-vcs/subversion")

    def exists(self, **kwargs):
        """Tests whether the repo actually exists"""
        return os.path.exists(os.path.join(self.repo.location, ".svn"))

    def new(self, **kwargs):
        if kwargs:
            self._kwargs(kwargs)
        # initial checkout
        svn_root = self.repo.sync_uri
        exitcode = portage.process.spawn_bash(
            "cd %s; exec svn co %s ."
            % (
                shlex.quote(self.repo.location),
                shlex.quote(svn_root),
            ),
            **self.spawn_kwargs,
        )
        if exitcode != os.EX_OK:
            msg = "!!! svn checkout error; exiting."
            self.logger(self.xterm_titles, msg)
            writemsg_level(msg + "\n", noiselevel=-1, level=logging.ERROR)
        return (exitcode, False)

    def update(self):
        """
        Internal function to update an existing SVN repository

        @return: tuple of return code (0=success), whether the cache
                needs to be updated
        @rtype: (int, bool)
        """

        exitcode = self._svn_upgrade()
        if exitcode != os.EX_OK:
            return (exitcode, False)

        # svn update
        exitcode = portage.process.spawn_bash(
            f"cd {shlex.quote(self.repo.location)}; exec svn update",
            **self.spawn_kwargs,
        )
        if exitcode != os.EX_OK:
            msg = "!!! svn update error; exiting."
            self.logger(self.xterm_titles, msg)
            writemsg_level(msg + "\n", noiselevel=-1, level=logging.ERROR)
        return (exitcode, False)

    def _svn_upgrade(self):
        """
        Internal function which performs an svn upgrade on the repo

        @return: tuple of return code (0=success), whether the cache
                needs to be updated
        @rtype: (int, bool)
        """
        exitcode = portage.process.spawn_bash(
            f"cd {shlex.quote(self.repo.location)}; exec svn upgrade",
            **self.spawn_kwargs,
        )
        if exitcode != os.EX_OK:
            msg = "!!! svn upgrade error; exiting."
            self.logger(self.xterm_titles, msg)
            writemsg_level(msg + "\n", noiselevel=-1, level=logging.ERROR)
        return exitcode



# test_PackageUseFile.py -- Portage Unit Testing Functionality
# Copyright 2007 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage import os
from portage.tests import TestCase
from portage.env.config import PackageUseFile
from tempfile import mkstemp


class PackageUseFileTestCase(TestCase):
    cpv = "sys-apps/portage"
    useflags = ["cdrom", "far", "boo", "flag", "blat"]

    def testPackageUseFile(self):
        """
        A simple test to ensure the load works properly
        """
        self.BuildFile()
        try:
            f = PackageUseFile(self.fname)
            f.load()
            for cpv, use in f.items():
                self.assertEqual(cpv, self.cpv)
                [flag for flag in use if self.assertTrue(flag in self.useflags)]
        finally:
            self.NukeFile()

    def BuildFile(self):
        fd, self.fname = mkstemp()
        f = os.fdopen(fd, "w")
        f.write(f"{self.cpv} {' '.join(self.useflags)}")
        f.close()

    def NukeFile(self):
        os.unlink(self.fname)



# Copyright 2017-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import functools
import random

from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop


class CallSoonFifoTestCase(TestCase):
    def testCallSoonFifo(self):
        event_loop = global_event_loop()
        inputs = [random.random() for index in range(10)]
        outputs = []
        finished = event_loop.create_future()

        def add_output(value):
            outputs.append(value)
            if len(outputs) == len(inputs):
                finished.set_result(True)

        for value in inputs:
            event_loop.call_soon(functools.partial(add_output, value))

        event_loop.run_until_complete(finished)
        self.assertEqual(inputs, outputs)



# Copyright 2013-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SlotConflictMaskUpdateTestCase(TestCase):
    def testBacktrackingGoodVersionFirst(self):
        """
        When backtracking due to slot conflicts, we masked the version that has been pulled
        in first. This is not always a good idea. Mask the highest version instead.
        """

        ebuilds = {
            "dev-libs/A-1": {"DEPEND": "=dev-libs/C-1 dev-libs/B"},
            "dev-libs/B-1": {"DEPEND": "=dev-libs/C-1"},
            "dev-libs/B-2": {"DEPEND": "=dev-libs/C-2"},
            "dev-libs/C-1": {},
            "dev-libs/C-2": {},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                mergelist=[
                    "dev-libs/C-1",
                    "dev-libs/B-1",
                    "dev-libs/A-1",
                ],
                success=True,
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds)

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2024  Alexey Gladkov <gladkov.alexey@gmail.com>

import os
import os.path
import logging
import zipfile
import shutil
import tempfile
import urllib.request

import portage
from portage.util import writemsg_level, writemsg_stdout
from portage.sync.syncbase import SyncBase


class ZipFile(SyncBase):
    """ZipFile sync module"""

    short_desc = "Perform sync operations on GitHub repositories"

    @staticmethod
    def name():
        return "ZipFile"

    def __init__(self):
        SyncBase.__init__(self, "emerge", ">=sys-apps/portage-2.3")

    def retrieve_head(self, **kwargs):
        """Get information about the checksum of the unpacked archive"""
        if kwargs:
            self._kwargs(kwargs)
        info = portage.grabdict(os.path.join(self.repo.location, ".info"))
        if "etag" in info:
            return (os.EX_OK, info["etag"][0])
        return (1, False)

    def _do_cmp(self, f1, f2):
        bufsize = 8 * 1024
        while True:
            b1 = f1.read(bufsize)
            b2 = f2.read(bufsize)
            if b1 != b2:
                return False
            if not b1:
                return True

    def sync(self, **kwargs):
        """Sync the repository"""
        if kwargs:
            self._kwargs(kwargs)

        req = urllib.request.Request(url=self.repo.sync_uri)

        info = portage.grabdict(os.path.join(self.repo.location, ".info"))
        if "etag" in info:
            req.add_header("If-None-Match", info["etag"][0])

        try:
            with urllib.request.urlopen(req) as response:
                with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
                    shutil.copyfileobj(response, tmp_file)

                zip_file = tmp_file.name
                etag = response.headers.get("etag")

        except urllib.error.HTTPError as resp:
            if resp.code == 304:
                writemsg_stdout(">>> The repository has not changed.\n", noiselevel=-1)
                return (os.EX_OK, False)

            writemsg_level(
                f"!!! Unable to obtain zip archive: {resp}\n",
                noiselevel=-1,
                level=logging.ERROR,
            )
            return (1, False)

        if not zipfile.is_zipfile(zip_file):
            msg = "!!! file is not a zip archive."
            self.logger(self.xterm_titles, msg)
            writemsg_level(msg + "\n", noiselevel=-1, level=logging.ERROR)

            os.unlink(zip_file)

            return (1, False)

        # Drop previous tree
        tempdir = tempfile.mkdtemp(prefix=".temp", dir=self.repo.location)
        tmpname = os.path.basename(tempdir)

        for name in os.listdir(self.repo.location):
            if name != tmpname:
                os.rename(
                    os.path.join(self.repo.location, name),
                    os.path.join(tempdir, name),
                )

        with zipfile.ZipFile(zip_file) as archive:
            strip_comp = 0

            for f in archive.namelist():
                f = os.path.normpath(f)
                if os.path.basename(f) == "profiles":
                    strip_comp = f.count("/")
                    break

            for n in archive.infolist():
                p = os.path.normpath(n.filename)

                if os.path.isabs(p):
                    continue

                parts = p.split("/")
                dstpath = os.path.join(self.repo.location, *parts[strip_comp:])

                if n.is_dir():
                    os.makedirs(dstpath, mode=0o755, exist_ok=True)
                    continue

                with archive.open(n) as srcfile:
                    prvpath = os.path.join(tempdir, *parts[strip_comp:])

                    if os.path.exists(prvpath):
                        with open(prvpath, "rb") as prvfile:
                            if self._do_cmp(prvfile, srcfile):
                                os.rename(prvpath, dstpath)
                                continue
                        srcfile.seek(0)

                    with open(dstpath, "wb") as dstfile:
                        shutil.copyfileobj(srcfile, dstfile)

        # Drop previous tree
        shutil.rmtree(tempdir)

        with open(os.path.join(self.repo.location, ".info"), "w") as infofile:
            if etag:
                infofile.write(f"etag {etag}\n")

        os.unlink(zip_file)

        return (os.EX_OK, True)



# Copyright 2005-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import portage
from portage import os
from portage.const import PRIVATE_PATH
from portage.util import grabdict, writedict


class CleanConfig:
    short_desc = "Discard any no longer installed configs from emerge's tracker list"

    @staticmethod
    def name():
        return "cleanconfmem"

    def __init__(self):
        self._root = portage.settings["ROOT"]
        self.target = os.path.join(portage.settings["EROOT"], PRIVATE_PATH, "config")

    def load_configlist(self):
        return grabdict(self.target)

    def check(self, **kwargs):
        onProgress = kwargs.get("onProgress", None)
        configs = self.load_configlist()
        messages = []
        maxval = len(configs)
        if onProgress:
            onProgress(maxval, 0)
            i = 0
        keys = sorted(configs)
        for config in keys:
            if not os.path.exists(config):
                messages.append(f"  {config}")
            if onProgress:
                onProgress(maxval, i + 1)
                i += 1
        msgs = self._format_output(messages)
        return (True, msgs)

    def fix(self, **kwargs):
        onProgress = kwargs.get("onProgress", None)
        configs = self.load_configlist()
        messages = []
        maxval = len(configs)
        if onProgress:
            onProgress(maxval, 0)
            i = 0

        root = self._root
        if root == "/":
            root = None
        modified = False
        for config in sorted(configs):
            if root is None:
                full_path = config
            else:
                full_path = os.path.join(root, config.lstrip(os.sep))
            if not os.path.exists(full_path):
                modified = True
                configs.pop(config)
                messages.append(f"  {config}")
            if onProgress:
                onProgress(maxval, i + 1)
                i += 1
        if modified:
            writedict(configs, self.target)
        msgs = self._format_output(messages, True)
        return (True, msgs)

    def _format_output(self, messages=[], cleaned=False):
        output = []
        if messages:
            output.append("Not Installed:")
            output += messages
            tot = "------------------------------------\n  Total %i Not installed"
            if cleaned:
                tot += " ...Cleaned"
            output.append(tot % len(messages))
        return output
